/*
 * ----------------------------------------------------------------------------

  Copyright (C) 2005 Thomas Pfeifer - http://thomaspfeifer.net/


	This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  US


Hardware:
  AT90S2313
  433 Mhz-Receiver with Data-Out connected to INT1
  FTDI232 connected to TxD

Bugs:
	will not work with negative temperatures

 * ----------------------------------------------------------------------------
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>


//####### CONFIG ########

//Note: if you use another xtal, you need to adjust the BITTIMEx values
#define F_CPU 4433619UL

#define BAUD 38400

#define MAXBITS 50 //max bits to receive

#define BITTIME0 68 //length of a 0 pulse
#define BITTIME1 21 //length of a 1 pulse
#define TOL 10      //length tolerance

//#define DEBUG //show me the bit-lengths

//#######################


//some useful stuff
#define cbi(reg, bit)   (reg &= ~_BV(bit))
#define sbi(reg, bit)   (reg |= _BV(bit))
#define uart(d) loop_until_bit_is_set(USR, UDRE); UDR = d;

//INT1 defines
#define INT_FALLING_EDGE sbi(MCUCR, ISC11); cbi(MCUCR, ISC10);
#define INT_RAISING_EDGE sbi(MCUCR, ISC11); sbi(MCUCR, ISC10);
#define INT_ENABLE sbi(GIMSK, INT1); // enable INT1
#define INT_DISABLE cbi(GIMSK, INT1); // disable INT1
#define INT_IS_RISING_MODE (MCUCR & (1<<ISC10))


volatile uint8_t	bittime[MAXBITS];
volatile uint8_t	bit[MAXBITS];
volatile uint8_t  bitpos=0;


#ifdef DEBUG
	static char hex[]="01234567890ABCDEF";
	void dec2Uart(uint8_t c) {
	  uart((c/100)+'0');
	  uart((c/10%10)+'0');
	  uart((c%10)+'0');
	}
#endif

//T0 Overflow
SIGNAL (SIG_OVERFLOW0) {
  INT_DISABLE;
  TCCR0=0; // Stop Timer

  uint8_t c;
  uint8_t ok=1;
  uint8_t i;

  for (i=0;i<bitpos;i++) {
 		c='-';

    //check for 1-bit
    if (bittime[i]>(BITTIME1-TOL) && bittime[i]<(BITTIME1+TOL)) {
    	bit[i]=1;
    	c='1';
    }

    //check for 0-bit
    if (bittime[i]>(BITTIME0-TOL) && bittime[i]<(BITTIME0+TOL)) {
    	bit[i]=0;
    	c='0';
    }

		if (c=='-') ok=0;
	  #ifdef DEBUG
      //writes A:B=C to the UART, where A is the Number of the bit,
      //B ist the pulse-length an C is the result.
      dec2Uart(i);
			uart(':');
			dec2Uart(bittime[i]);
			uart('=');
			uart(c);
    	uart('\r');
    #endif

	}

  for (i=1;i<bitpos;i+=2) {
  	uart(bit[i]+'0');
  }
  uart('\r');
  for (i=0;i<bitpos;i+=2) {
  	uart(bit[i]+'0');
  }
  uart('\r');
	#ifdef DEBUG
  	uart(13);
  #endif


  if (bitpos==44 && ok) {

  	// decode temperature if # of bits is correct and all bits are
		// successfully recognized.
  	uart((8*bit[20]+4*bit[21]+2*bit[22]+1*bit[23])+'0'-5); //FIXME: will not work with negative t ;-)
	  uart((8*bit[24]+4*bit[25]+2*bit[26]+1*bit[27])+'0');
  	uart(',');
  	uart((8*bit[28]+4*bit[29]+2*bit[30]+1*bit[31])+'0');
  	uart('');
  	uart('C');
  	uart(13);

  } else {

  	// write x if decoding has failed
    uart('x');
    uart(13);
  }

	bitpos=0;
	INT_RAISING_EDGE;
	INT_ENABLE;
}

//INT1
SIGNAL (SIG_INTERRUPT1) {
	INT_DISABLE;

	if (INT_IS_RISING_MODE) {
		INT_FALLING_EDGE;
	} else {
	  if (bitpos<(MAXBITS-1)) {
    	bittime[bitpos++]=TCNT0;
  	}
		INT_RAISING_EDGE;
	}

	INT_ENABLE;
	sbi(GIMSK, INT1); // disable ext. int.

	TCNT0=0;
	TCCR0=_BV(CS01) | _BV(CS00);
}


int main (void) {

	DDRB  = 0b00000000;                  // 1=Output, 0=Input
	PORTB = 0b00000000;                  // activate internal pull-up

	DDRD  = 0b00000000;                  // 1=Output, 0=Input
	PORTD = 0b00000000;                  // activate internal pull-up

  // UART
	sbi(UCR,TXEN); //enable Uart-TX
	UBRR=(F_CPU / (BAUD * 16L) - 1); //set Baudrate

  // TIMER0
  sbi(TIMSK, TOIE0); // enable T0 Overflow


  sei();  //enable Interrupts


	//GIFR = (1<<INTF1);
	INT_ENABLE;
  INT_RAISING_EDGE;



	for (;;) {
	}
}




